"use strict";
var __assign = (this && this.__assign) || Object.assign || function(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
            t[p] = s[p];
    }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
/* Copyright IBM Corp. 2017 */
var fs_extra_1 = require("fs-extra");
var graceful_fs_1 = require("graceful-fs");
var path_1 = require("path");
var rxjs_1 = require("rxjs");
var operators_1 = require("rxjs/operators");
var verror_1 = require("verror");
var rx_utils_1 = require("./rx.utils");
var rx_walk_1 = require("./rx.walk");
/**
 * Creates a single directory
 *
 * @param aPath     the path to create
 *
 * @return observable with the result
 */
function _rxMkDir(aPath) {
    return rxjs_1.Observable.create(function (observer) {
        // create the path
        graceful_fs_1.mkdir(aPath, function (mkdirError) {
            if (mkdirError) {
                // quick path to check for an existing directory
                if (mkdirError.code === 'EEXIST') {
                    // success
                    observer.next(aPath);
                    observer.complete();
                }
                else {
                    // test if the directory exists
                    fs_extra_1.stat(aPath, function (statError, stat) {
                        // in case of a stat error, bail our
                        if (statError) {
                            observer.error(new verror_1.VError(statError, 'Unable to create directory [%s].', aPath));
                        }
                        else 
                        // check if the directory exists
                        if (stat.isDirectory()) {
                            // success
                            observer.next(aPath);
                            observer.complete();
                        }
                        else {
                            // bail out
                            observer.error(new verror_1.VError(mkdirError, 'Unable to create directory [%s].', aPath));
                        }
                    });
                }
            }
            else {
                observer.next(aPath);
                observer.complete();
            }
        });
    });
}
function _rxMkDirs() {
    // directory map
    var dirMap = {};
    function _mkdir(aDir) {
        // test if we know the directory
        return dirMap[aDir] || (dirMap[aDir] = rx_utils_1.rxCacheSingle(_rxMkDir(aDir)));
    }
    function _mkdirp(aDir) {
        // check if we know the directory
        var res = dirMap[aDir];
        if (!res) {
            // construct the parent directory
            var parsed = path_1.parse(aDir);
            // test for the root directory
            if (parsed.root === parsed.dir) {
                // do not recurse
                res = rxjs_1.of(aDir);
            }
            else {
                // construct the parent and attach
                res = _mkdirp(parsed.dir).pipe(operators_1.mergeMap(function () { return _mkdir(aDir); }), operators_1.catchError(function (error) { return rxjs_1.throwError(new verror_1.VError(error, 'Unable to recursively create directory [%s].', aDir)); }));
            }
        }
        // ok
        return res;
    }
    return _mkdirp;
}
exports.rxMkDirs = _rxMkDirs;
/**
 * Copies one directory into the other and issues the names of the copied files
 *
 * @param aSrc  the source directory
 * @param aDst  the target directory
 * @param aOverride true if we want to override, else false
 *
 * @return observable of the copied files
 */
function _doCopyDir(aSrc, aDst, aMkdirp, aOverride) {
    // create the target
    var rxDstDir = aMkdirp(aDst);
    // children
    var rxChildren = rx_walk_1.rxReaddir(aSrc).pipe(operators_1.mergeMap(function (children) { return rxjs_1.from(children); }));
    var opFileInfo = function (rxChild) { return rxChild.pipe(operators_1.map(function (child) { return ({ child: child, src: path_1.join(aSrc, child), dst: path_1.join(aDst, child) }); })); };
    var opStats = function (rxInfo) { return rxInfo.pipe(operators_1.mergeMap(function (info) { return rx_walk_1.rxStats(info.src).pipe(operators_1.map(function (stats) { return (__assign({}, info, { stats: stats })); })); })); };
    var opFileInfoWithStats = function (rxFileInfo) { return rxFileInfo.pipe(opFileInfo, opStats); };
    var rxFileInfoWithStats = rxChildren.pipe(opFileInfoWithStats, operators_1.share());
    // split the children into directories and files
    var rxDirs = rxFileInfoWithStats.pipe(operators_1.filter(function (info) { return info.stats.isDirectory(); }));
    var rxFiles = rxFileInfoWithStats.pipe(operators_1.filter(function (info) { return info.stats.isFile(); }));
    // copy the files
    var rxCopyFiles = rxFiles.pipe(operators_1.mergeMap(function (info) { return rx_walk_1.rxCopyFile(info.src, info.dst); }));
    // copy the directory
    var rxCopyDir = rxDstDir.pipe(operators_1.mergeMap(function () { return rxDirs.pipe(operators_1.mergeMap(function (info) { return _doCopyDir(info.src, info.dst, aMkdirp, aOverride); })); }));
    // combine both
    return rxjs_1.merge(rxCopyDir, rxCopyFiles);
}
/**
 * Copies a set of source files to a destination directory
 *
 * @param aSources  the list of source files
 * @param aDst  the target directory
 *
 * @return the list of copied filed
 */
function _copyFiles(aSources, aDst) {
    // mkdir hook
    var _mkdirp = _rxMkDirs();
    /**
     * Copies a single file from source to destination
     *
     * @param aSrcFile  the source file
     * @param aDstFile  the destination file
     *
     * @return observable of the copy result
     */
    function _copyFile(aSrcFile, aDstFile) {
        // copy
        var parsed = path_1.parse(aDstFile);
        // do
        return _mkdirp(parsed.dir).pipe(operators_1.mergeMapTo(rx_walk_1.rxCopyFileStream(aSrcFile, aDstFile)), operators_1.catchError(function (error) { return rxjs_1.throwError(new verror_1.VError(error, 'Unable to copy file [%s] to [%s].', aSrcFile, aDstFile)); }));
    }
    // execute the copy process
    return aSources.pipe(operators_1.filter(function (src) { return !src.stats.isDirectory(); }), operators_1.mergeMap(function (src) { return _copyFile(src.absPath, path_1.join(aDst, src.relPath)); }));
}
exports.rxCopyFiles = _copyFiles;
// copy
var _rxCopyDir = function (aSrc, aDst, aOverride) {
    return _doCopyDir(aSrc, aDst, _rxMkDirs(), !!aOverride).pipe(operators_1.catchError(function (copyError) { return rxjs_1.throwError(new verror_1.VError(copyError, 'Unable to copy [%s] to [%s].', aSrc, aDst)); }));
};
exports.rxCopyDir = _rxCopyDir;
