/* Copyright IBM Corp. 2017 */
import { readFile, writeFile } from 'graceful-fs';
import { extname } from 'path';
import { Observable, Observer } from 'rxjs';
import { filter, map, mergeMap } from 'rxjs/operators';

import { firstOf } from '../utils/file.utils';
import { CMD_ADD, TYPE_COPYRIGHT } from './../constants';
import { Options } from './../options';
import { FileDescriptor, walkFiles } from './../utils/rx.walk';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';

const EXTENSIONS = ['.ts', '.css'];

const COPYRIGHT = '/* Copyright IBM Corp. 2017 */\n';

function _addCopyright(aFile: FileDescriptor): Observable<FileDescriptor> {
    // new observable
    return Observable.create((observer: Observer<FileDescriptor>) => {
        // load the file
        readFile(aFile.absPath, 'utf-8', (errRead, data) => {
            if (errRead) {
                observer.error(errRead);
            } else
                // test if we need a copyright
                if (data.indexOf(COPYRIGHT) !== 0) {
                    // add the copyright
                    const newData = COPYRIGHT + data;
                    writeFile(aFile.absPath, newData, errWrite => {
                        if (errWrite) {
                            observer.error(errWrite);
                        } else {
                            observer.next(aFile);
                            observer.complete();
                        }
                    });
                } else {
                    observer.complete();
                }
        });
    });
}

function _filterExtension(aFile: FileDescriptor): boolean {
    // decode the file type
    const ext = extname(aFile.absPath);
    // check
    return EXTENSIONS.indexOf(ext) >= 0;
}

export class AddCopyrightCommand extends AbstractCommand implements Command {

    static readonly TYPE = TYPE_COPYRIGHT;

    static readonly COMMAND = CMD_ADD;

    constructor(private options: Options) {
        super(options);
    }

    process(): Observable<string> {
        const opt = this.options;
        // load the compiler options
        const rootDir = firstOf(opt.src) || '.';
        // iterate over all eligible files
        return walkFiles(rootDir).pipe(
            filter(_filterExtension),
            mergeMap(_addCopyright),
            map(file => file.absPath)
        );
    }
}
