/* Copyright IBM Corp. 2017 */
import { cloneDeep } from 'lodash';
import { join } from 'path';
import { Observable, of, Subject } from 'rxjs';
import { mergeMap } from 'rxjs/operators';

import { CMD_CREATE, OPTION_DATA, OPTION_NAME, OPTION_VERBOSE } from './../constants';
import { Options } from './../options';
import { assertNotNull } from './../utils/assert';
import { canonicalizeJSON } from './../utils/json';
import { writeJsonFileSafe } from './../utils/rx.file';
import { rxMkdirp } from './../utils/rx.walk';
import { SiteNode, WchTools } from './../utils/wchtools';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';
import { HELP_APPLICATION_NAME, HELP_DATA_DIRECTORY, HELP_VERBOSE } from './help';

function _byPosition(aLeft: any, aRight: any): number {
    const leftPos = (aLeft && aLeft.position) ? aLeft.position : Number.MAX_SAFE_INTEGER,
        rightPos = (aRight && aRight.position) ? aRight.position : Number.MAX_SAFE_INTEGER;
    // compare
    return (leftPos < rightPos) ? -1 : (leftPos > rightPos) ? +1 : 0;
}

function _buildSiteChildren(aRootPath: string, aChildren: SiteNode[] | null | undefined): any[] {
    const result = aChildren ? aChildren.map(node => {
        const res = cloneDeep(node.data);
        // build the route
        const route = aRootPath + encodeURIComponent(res.segment) + '/';
        res.route = route;
        // recurse
        res.children = _buildSiteChildren(route, node.children);
        return res;
    }) : [];
    // sort
    return result.sort(_byPosition);
}

function _writeSite(aSiteName: string, aAppAssetDir: string, aSite: any, aLogger: Subject<string>): Observable<string> {
    // some assertions
    assertNotNull(aSiteName, 'siteName');
    assertNotNull(aAppAssetDir, 'assetDir');
    assertNotNull(aSite, 'site');
    assertNotNull(aLogger, 'logger');
    // build the target folder
    const dirName = join(aAppAssetDir, 'delivery', 'v1', 'rendering', 'sites');
    const fileName = join(dirName, aSiteName + '.json');
    // log this
    aLogger.next(`Writing site feed to [${fileName}] ...`);
    // construct the target directory
    return rxMkdirp(dirName).pipe(
        mergeMap(() => writeJsonFileSafe(fileName, aSite, true))
    );
}

function _buildSite(aSiteName: string, aTools: WchTools): Observable<any> {
    // the result
    let result: any = {};
    // access the site
    const root: SiteNode = aTools.sites[aSiteName];
    if (root) {
        // start with the root
        result = cloneDeep(root.data);
        result.pages = _buildSiteChildren('/', root.children);
    }
    // done
    return of(canonicalizeJSON(result));
}

/**
 * Generates angular components based on type and layout mapping definitions
 *
 * create site --data DATA_FOLDER --name APP_NAME
 */
export class CreateSiteCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'site';

    static readonly COMMAND = CMD_CREATE;

    constructor(private options: Options) {
        super(options);
    }

    printHelp(): void {
        // log the command line
        console.log(CreateSiteCommand.COMMAND, CreateSiteCommand.TYPE,
            `--${OPTION_NAME} <NAME>`,
            `[--${OPTION_DATA} <DATA>]`,
            `[--${OPTION_VERBOSE}]`);
        // log the parameters
        console.log(`--${OPTION_NAME}`, ':', HELP_APPLICATION_NAME);
        console.log(`--${OPTION_DATA}`, ':', HELP_DATA_DIRECTORY);
        console.log(`--${OPTION_VERBOSE}`, ':', HELP_VERBOSE);
    }

    process(): Observable<string> {
        // logger
        const logger = this.getLogger();
        // data directory
        // current options
        const opts = this.options;
        // site name
        const siteName = 'default';
        // the directories
        const appName = opts.name || this.getAppName();
        const appAssetDir = this.getAppAssetDir(appName);
        // the tools
        const rxWchTools = this.getWchTools();
        // the site
        const rxSite = rxWchTools.pipe(
            mergeMap(aTools => _buildSite(siteName, aTools)),
            mergeMap(aSite => _writeSite(siteName, appAssetDir, aSite, logger))
        );

        // ok
        return rxSite;
    }
}
