/* Copyright IBM Corp. 2017 */
import { getType } from 'mime';
import { empty, Observable, of, Subject } from 'rxjs';
import { filter, mergeMap, tap } from 'rxjs/operators';

import { CMD_NORMALIZE, OPTION_DATA, OPTION_VERBOSE } from './../constants';
import { canonicalizeJSON } from './../utils/json';
import { readFile as rxReadFile, walkFiles as rxWalkFiles, writeFile as rxWriteFile } from './../utils/rx.walk';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';
import { HELP_VERBOSE } from './help';

function _normalizeSingleJson(aJson: any): any {
    return canonicalizeJSON(aJson);
}

function _normalizeSingleJsonFile(aFileName: string, aLogger: Subject<string>): Observable<string> {
    // read file
    return of(aFileName).pipe(
        tap(name => aLogger.next(`Reading [${name}].`)),
        mergeMap(name => rxReadFile(aFileName)),
        mergeMap(data => {
            const normalized = JSON.stringify(_normalizeSingleJson(JSON.parse(data)), undefined, 2);
            return (data !== normalized) ? rxWriteFile(aFileName, normalized).pipe(
                tap(name => aLogger.next(`Written [${name}].`))) : empty();
        })
    );
}

function _normalizeAll(aDir: string, aLogger: Subject<string>): Observable<string> {
    // walk all JSON files
    return rxWalkFiles(aDir).pipe(
        filter(file => getType(file.absPath) === 'application/json'),
        mergeMap(file => _normalizeSingleJsonFile(file.absPath, aLogger))
    );
}

export class NormalizeJsonCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'json';
    static readonly COMMAND = CMD_NORMALIZE;

    printHelp(): void {
        // log the command line
        console.log(NormalizeJsonCommand.COMMAND, NormalizeJsonCommand.TYPE,
            `[--${OPTION_DATA} <DATA>]`,
            `[--${OPTION_VERBOSE}]`);
        // log the parameters
        console.log(`--${OPTION_DATA}`, ':', OPTION_DATA);
        console.log(`--${OPTION_VERBOSE}`, ':', HELP_VERBOSE);
    }

    process(): Observable<string> {
        // start with the data directory
        const dataDir = this.getDataDir();
        // logging
        const logger = this.getLogger();
        // process
        return _normalizeAll(dataDir, logger);
    }
}
