"use strict";
exports.__esModule = true;
var lodash_1 = require("lodash");
var number_to_words_1 = require("number-to-words");
var path_1 = require("path");
var operators_1 = require("rxjs/operators");
var transliteration_1 = require("./../utils/transliteration");
var rx_dir_1 = require("./rx.dir");
function _camelCase(aValue) {
    return transliteration_1.camelCase(aValue);
}
function _kebabCase(aValue) {
    return transliteration_1.kebabCase(aValue);
}
var TYPE_SUFFIX = 'Type';
var ELEMENT_SUFFIX = 'Element';
var ELEMENTS_PATH = 'elements';
var SINGLE_PREFIX = 'Single';
var MULTI_PREFIX = 'Multi';
var LEADING_DIGIT = /^(\d*)/g;
function _createTypeComponentName(aName) {
    // trim
    var trimmedName = aName.trim();
    var fixed = trimmedName.replace(LEADING_DIGIT, function (m, p) { return (p && (p.length > 0)) ? number_to_words_1.toWords(p) + ' ' : ''; });
    /** check if the layout component starts with a number, in this case
     * convert the number to a string */
    var name = lodash_1.upperFirst(_camelCase(fixed));
    return name.endsWith(TYPE_SUFFIX) ? name : name + TYPE_SUFFIX;
}
function _createElementComponentName(aName) {
    // trim
    var trimmedName = aName.trim();
    var fixed = trimmedName.replace(LEADING_DIGIT, function (m, p) { return (p && (p.length > 0)) ? number_to_words_1.toWords(p) + ' ' : ''; });
    /** check if the layout component starts with a number, in this case
     * convert the number to a string */
    var name = lodash_1.upperFirst(_camelCase(fixed));
    return name.endsWith(TYPE_SUFFIX) ? name.substr(0, name.length - TYPE_SUFFIX.length) + ELEMENT_SUFFIX : name + ELEMENT_SUFFIX;
}
function _getPathForType(aType, aOptions) {
    // name of the layout folder segment
    var typeName = _createTypeComponentName(aType.name);
    if (typeName.endsWith(TYPE_SUFFIX)) {
        typeName = typeName.substr(0, typeName.length - TYPE_SUFFIX.length);
    }
    // layout folder
    var typeFolder = _kebabCase(typeName);
    // access the type properties
    var tags = aType.tags || [];
    // convert tags into segments
    var segments = aOptions.flat ? [] : tags.map(_kebabCase).sort();
    segments.push(typeFolder);
    // return this path
    return segments.join(path_1.sep);
}
function _getTypePath(aType, aOptions) {
    // join
    return path_1.join(ELEMENTS_PATH, _getPathForType(aType, aOptions));
}
/**
 * Constructs the relevant information for a components class
 *
 * @param aDir      base directory
 * @param aType     type object
 * @param aOptions  options
 * @param aLogger   logger
 *
 * @return the component object
 */
function _initTypeClass(aDir, aType, aOptions, aLogger) {
    // check for the path suffix
    var typePath = _getTypePath(aType, aOptions);
    var type = aType;
    var typeName = type.name;
    // component names
    var typeClass = _createTypeComponentName(typeName);
    var simpleTypeClass = typeClass;
    var typeElementClass = typeClass.endsWith(TYPE_SUFFIX) ? typeClass.substr(0, typeClass.length - TYPE_SUFFIX.length) : typeClass;
    var baseElementClass = _createElementComponentName(typeName);
    var singleElementClass = SINGLE_PREFIX + baseElementClass;
    var multiElementClass = MULTI_PREFIX + baseElementClass;
    // references
    var typeElementRef = _camelCase(typeClass);
    // file names
    var typeElementFile = typeElementRef + '.ts';
    // the base path for the layout
    var folder = path_1.normalize(path_1.join(aDir, typePath));
    var result = {
        type: type, folder: folder,
        simpleTypeClass: simpleTypeClass,
        typeElementClass: typeElementClass,
        baseElementClass: baseElementClass,
        singleElementClass: singleElementClass,
        multiElementClass: multiElementClass,
        typeElementFile: typeElementFile,
        typeElementRef: typeElementRef
    };
    // log this
    aLogger.next("Element: " + typeName + ", ElementClass: " + JSON.stringify(typeClass) + ", Folder: " + folder);
    // ok
    return result;
}
/**
 * Finds the directory tha the component should be created in. Tests
 * each potential source directory for an existing component file. If non exists, uses the
 * first source directory
 *
 * @param aType        type
 * @param aSources     sources directory
 *
 * @return the selected source directory
 */
function _getTypesDir(aType, aSources, aOptions, aMkdirp, aLogger) {
    // check for the path suffix
    var typePath = _getTypePath(aType, aOptions);
    var className = _createTypeComponentName(aType.name);
    var relPath = path_1.join(typePath, _camelCase(className) + '.ts');
    // map each source to a check
    return rx_dir_1.rxFindDir(aSources, relPath, aMkdirp).pipe(operators_1.tap(function (dir) { return aLogger.next("Selected type source directory [" + dir + "] for type [" + className + "]."); }));
}
function _createTypeClass(aType, aSources, aOptions, aMkdirp, aLogger) {
    // locate the component directory
    return _getTypesDir(aType, aSources, aOptions, aMkdirp, aLogger).pipe(operators_1.first(), operators_1.map(function (dir) { return _initTypeClass(dir, aType, aOptions, aLogger); }), operators_1.mergeMap(function (clz) { return aMkdirp(clz.folder).pipe(operators_1.mapTo(clz)); }), operators_1.shareReplay());
}
var TypeRegistry = /** @class */ (function () {
    function TypeRegistry(aBaseFolders, aOptions, aMkdirp, aLogger) {
        this.aBaseFolders = aBaseFolders;
        this.aOptions = aOptions;
        this.aMkdirp = aMkdirp;
        this.aLogger = aLogger;
        this.typeRegistry = {};
        this.typeMap = {};
    }
    TypeRegistry.prototype.registerType = function (aType) {
        var type = aType;
        var id = type.id;
        if (this.typeMap[id]) {
            // do not create a type, twice
            return false;
        }
        // register
        this.typeMap[id] = type;
        // ok
        return true;
    };
    TypeRegistry.prototype.findTypeClass = function (aType) {
        // the name
        var typeId = aType.id;
        // check if we are already searching
        var rxRunning = this.typeRegistry[typeId];
        if (!rxRunning) {
            // try to find the folder
            rxRunning = this.typeRegistry[typeId] =
                _createTypeClass(aType, this.aBaseFolders, this.aOptions, this.aMkdirp, this.aLogger);
        }
        // ok
        return rxRunning;
    };
    return TypeRegistry;
}());
exports.TypeRegistry = TypeRegistry;
